from __future__ import annotations

import warnings
from typing import Any

import pandas as pd
from airflow.decorators.base import get_unique_task_id
from airflow.hooks.base import BaseHook
from airflow.models.xcom_arg import XComArg

from astro.airflow.datasets import kwargs_with_datasets
from astro.constants import DEFAULT_CHUNK_SIZE, ColumnCapitalization, LoadExistStrategy
from astro.databases import create_database
from astro.databases.base import BaseDatabase
from astro.dataframes.pandas import PandasDataframe
from astro.files import File, resolve_file_path_pattern
from astro.options import LoadOptions, LoadOptionsList
from astro.settings import LOAD_FILE_ENABLE_NATIVE_FALLBACK
from astro.sql.operators.base_operator import AstroSQLBaseOperator
from astro.table import BaseTable
from astro.utils.compat.typing import Context


class LoadFileOperator(AstroSQLBaseOperator):
    """Load S3/local file into either a database or a pandas dataframe

    :param input_file: File path and conn_id for object stores
    :param output_table: Table to create
    :param ndjson_normalize_sep: separator used to normalize nested ndjson.
    :param chunk_size: Specify the number of records in each batch to be written at a time.
    :param if_exists: Overwrite file if exists. Default False.
    :param use_native_support: Use native support for data transfer if available on the destination.
    :param native_support_kwargs: kwargs to be used by method involved in native support flow
    :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
    :param enable_native_fallback: Use enable_native_fallback=True to fall back to default transfer

    :return: If ``output_table`` is passed this operator returns a Table object. If not
        passed, returns a dataframe.
    """

    template_fields = ("output_table", "input_file")

    def __init__(
        self,
        input_file: File,
        output_table: BaseTable | None = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        if_exists: LoadExistStrategy = "replace",
        ndjson_normalize_sep: str = "_",
        use_native_support: bool = True,
        native_support_kwargs: dict | None = None,
        load_options: list[LoadOptions] | None = None,
        columns_names_capitalization: ColumnCapitalization = "original",
        enable_native_fallback: bool | None = LOAD_FILE_ENABLE_NATIVE_FALLBACK,
        **kwargs,
    ) -> None:
        kwargs.setdefault("task_id", get_unique_task_id("load_file"))
        super().__init__(
            **kwargs_with_datasets(
                kwargs=kwargs,
                input_datasets=input_file,
                output_datasets=output_table,
            )
        )
        if native_support_kwargs:
            warnings.warn(
                """`load_options` will replace `native_support_kwargs` parameter in astro-sdk-python>=1.5.0. Please
                use `load_options` parameter instead.""",
                DeprecationWarning,
                stacklevel=2,
            )
        self.output_table = output_table
        self.input_file = input_file
        self.input_file.load_options = load_options
        self.chunk_size = chunk_size
        self.kwargs = kwargs
        self.if_exists = if_exists
        self.ndjson_normalize_sep = ndjson_normalize_sep
        self.normalize_config: dict[str, str] = {}
        self.use_native_support = use_native_support
        self.native_support_kwargs: dict[str, Any] = native_support_kwargs or {}
        self.columns_names_capitalization = columns_names_capitalization
        self.enable_native_fallback = enable_native_fallback
        self.load_options_list = LoadOptionsList(load_options)

    def execute(self, context: Context) -> BaseTable | File:  # skipcq: PYL-W0613
        """
        Load an existing dataset from a supported file into a SQL table or a Dataframe.
        """
        if self.input_file.conn_id:
            check_if_connection_exists(self.input_file.conn_id)

        # TODO: remove pushing to XCom once we update the airflow version.
        if self.output_table:
            context["ti"].xcom_push(key="output_table_conn_id", value=str(self.output_table.conn_id))
            context["ti"].xcom_push(key="output_table_name", value=str(self.output_table.name))
        return self.load_data(input_file=self.input_file, context=context)

    def load_data(self, input_file: File, context: Context) -> BaseTable | pd.DataFrame:

        self.log.info("Loading %s into %s ...", self.input_file.path, self.output_table)
        if self.output_table:
            return self.load_data_to_table(input_file, context)
        else:
            return self.load_data_to_dataframe(input_file)

    def load_data_to_table(self, input_file: File, context: Context) -> BaseTable:
        """
        Loads csv/parquet table from local/S3/GCS with Pandas.
        Infers SQL database type based on connection then loads table to db.
        """
        if not isinstance(self.output_table, BaseTable):
            raise ValueError("Please pass a valid Table instance in 'output_table' parameter")
        database = create_database(
            self.output_table.conn_id, self.output_table, load_options_list=self.load_options_list
        )
        self.output_table = database.populate_table_metadata(self.output_table)
        normalize_config = self._populate_normalize_config(
            ndjson_normalize_sep=self.ndjson_normalize_sep,
            database=database,
        )
        database.load_file_to_table(
            input_file=input_file,
            normalize_config=normalize_config,
            output_table=self.output_table,
            if_exists=self.if_exists,
            chunk_size=self.chunk_size,
            use_native_support=self.use_native_support,
            native_support_kwargs=self.native_support_kwargs,
            columns_names_capitalization=self.columns_names_capitalization,
            enable_native_fallback=self.enable_native_fallback,
            databricks_job_name=f"Load data {self.dag_id}_{self.task_id}",
        )
        self.log.info("Completed loading the data into %s.", self.output_table)
        return self.output_table

    def load_data_to_dataframe(self, input_file: File) -> pd.DataFrame | None:
        """
        Loads csv/parquet file from local/S3/GCS with Pandas. Returns dataframe as no
        SQL table was specified
        """
        df = None
        for file in resolve_file_path_pattern(
            input_file.path,
            input_file.conn_id,
            normalize_config=self.normalize_config,
            filetype=input_file.type.name,
            load_options=input_file.load_options,
        ):
            if isinstance(df, pd.DataFrame):
                df = pd.concat(
                    [
                        df,
                        file.export_to_dataframe(
                            columns_names_capitalization=self.columns_names_capitalization
                        ),
                    ],
                    ignore_index=True,
                )
            else:
                df = file.export_to_dataframe(
                    columns_names_capitalization=self.columns_names_capitalization,
                )

        if not isinstance(df, PandasDataframe):
            df = PandasDataframe.from_pandas_df(df)

        self.log.info("Completed loading the data into dataframe.")
        return df

    @staticmethod
    def _populate_normalize_config(
        database: BaseDatabase,
        ndjson_normalize_sep: str = "_",
    ) -> dict[str, str]:
        """
        Validate pandas json_normalize() parameter for databases, since default params result in
        invalid column name. Default parameter result in the columns name containing '.' char.

        :param ndjson_normalize_sep: separator used to normalize nested ndjson.
            https://pandas.pydata.org/docs/reference/api/pandas.json_normalize.html
        :param database: supported database
        """

        def replace_illegal_columns_chars(char: str, database: BaseDatabase) -> str:
            index = (
                database.illegal_column_name_chars.index(char)
                if char in database.illegal_column_name_chars
                else None
            )
            if index is not None:
                return str(database.illegal_column_name_chars_replacement[index])
            else:
                return str(char)

        normalize_config: dict[str, Any] = {
            "meta_prefix": ndjson_normalize_sep,
            "record_prefix": ndjson_normalize_sep,
            "sep": ndjson_normalize_sep,
        }
        normalize_config["meta_prefix"] = replace_illegal_columns_chars(
            normalize_config["meta_prefix"], database
        )
        normalize_config["record_prefix"] = replace_illegal_columns_chars(
            normalize_config["record_prefix"], database
        )
        normalize_config["sep"] = replace_illegal_columns_chars(normalize_config["sep"], database)

        return normalize_config

    def get_openlineage_facets_on_complete(self, task_instance):  # skipcq: PYL-W0613
        """
        Returns the lineage data
        """
        from astro.lineage import (
            BaseFacet,
            DataSourceDatasetFacet,
            OpenlineageDataset,
            OperatorLineage,
            SchemaDatasetFacet,
            SchemaField,
        )
        from astro.lineage.facets import InputFileDatasetFacet, InputFileFacet, OutputDatabaseDatasetFacet

        # if the input_file is a folder or pattern, it needs to be resolved to
        # list the files
        input_files = resolve_file_path_pattern(
            self.input_file.path,
            self.input_file.conn_id,
            normalize_config={},
            filetype=self.input_file.type.name,
            load_options=self.input_file.load_options,
        )

        input_uri = (
            f"{self.input_file.openlineage_dataset_namespace}{self.input_file.openlineage_dataset_name}"
        )
        input_dataset: list[OpenlineageDataset] = [
            OpenlineageDataset(
                namespace=self.input_file.openlineage_dataset_namespace,
                name=self.input_file.openlineage_dataset_name,
                facets={
                    "input_file_facet": InputFileDatasetFacet(
                        is_pattern=self.input_file.is_pattern(),
                        number_of_files=len(input_files),
                        files=[
                            InputFileFacet(
                                filepath=file.path,
                                file_size=file.size,
                                file_type=file.type.name,
                            )
                            for file in input_files
                        ],
                    ),
                    "dataSource": DataSourceDatasetFacet(
                        name=self.input_file.openlineage_dataset_name, uri=input_uri
                    ),
                },
            )
        ]

        output_dataset: list[OpenlineageDataset] = []
        if self.output_table is not None and self.output_table.openlineage_emit_temp_table_event():
            # TODO: remove pushing to XCom once we update the airflow version.
            self.output_table.conn_id = task_instance.xcom_pull(
                task_ids=task_instance.task_id, key="output_table_conn_id"
            )
            self.output_table.name = task_instance.xcom_pull(
                task_ids=task_instance.task_id, key="output_table_name"
            )
            output_dataset = [
                OpenlineageDataset(
                    namespace=self.output_table.openlineage_dataset_namespace(),
                    name=self.output_table.openlineage_dataset_name(),
                    facets={
                        "output_database_facet": OutputDatabaseDatasetFacet(
                            metadata=self.output_table.metadata,
                            columns=self.output_table.columns,
                            schema=self.output_table.sqlalchemy_metadata.schema,
                            used_native_path=self.use_native_support,
                            enabled_native_fallback=self.enable_native_fallback,
                            native_support_arguments=self.native_support_kwargs,
                        ),
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaField(
                                    name=self.output_table.metadata.schema,
                                    type=self.output_table.metadata.database,
                                )
                            ]
                        ),
                        "dataSource": DataSourceDatasetFacet(
                            name=self.output_table.name, uri=self.output_table.openlineage_dataset_uri()
                        ),
                    },
                )
            ]

        run_facets: dict[str, BaseFacet] = {}
        job_facets: dict[str, BaseFacet] = {}

        return OperatorLineage(
            inputs=input_dataset, outputs=output_dataset, run_facets=run_facets, job_facets=job_facets
        )


def load_file(
    input_file: File,
    output_table: BaseTable | None = None,
    task_id: str | None = None,
    if_exists: LoadExistStrategy = "replace",
    ndjson_normalize_sep: str = "_",
    use_native_support: bool = True,
    native_support_kwargs: dict | None = None,
    columns_names_capitalization: ColumnCapitalization = "original",
    enable_native_fallback: bool | None = True,
    load_options: list[LoadOptions] | None = None,
    **kwargs: Any,
) -> XComArg:
    """Load a file or bucket into either a SQL table or a pandas dataframe.

    :param input_file: File path and conn_id for object stores
    :param output_table: Table to create
    :param task_id: task id, optional
    :param if_exists: default override an existing Table. Options: fail, replace, append
    :param ndjson_normalize_sep: separator used to normalize nested ndjson.
        ex - ``{"a": {"b":"c"}}`` will result in: ``column - "a_b"`` where ``ndjson_normalize_sep = "_"``
    :param use_native_support: Use native support for data transfer if available on the destination.
    :param native_support_kwargs: kwargs to be used by method involved in native support flow
    :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
        in the resulting dataframe
    :param enable_native_fallback: Use enable_native_fallback=True to fall back to default transfer
    :param load_options: load options while reading and loading file
    """

    # Note - using path for task id is causing issues as it's a pattern and
    # contain chars like - ?, * etc. Which are not acceptable as task id.
    task_id = task_id if task_id is not None else get_unique_task_id("load_file")

    if native_support_kwargs:
        warnings.warn(
            """`load_options` will replace `native_support_kwargs` parameter in astro-sdk-python>=1.5.0. Please use
            `load_options` parameter instead.""",
            DeprecationWarning,
            stacklevel=2,
        )

    return LoadFileOperator(
        task_id=task_id,
        input_file=input_file,
        output_table=output_table,
        if_exists=if_exists,
        ndjson_normalize_sep=ndjson_normalize_sep,
        use_native_support=use_native_support,
        native_support_kwargs=native_support_kwargs,
        columns_names_capitalization=columns_names_capitalization,
        enable_native_fallback=enable_native_fallback,
        load_options=load_options,
        **kwargs,
    ).output


def check_if_connection_exists(conn_id: str) -> bool:
    """
    Given an Airflow connection ID, identify if it exists.
    Return True if it does or raise an AirflowNotFoundException exception if it does not.

    :param conn_id: Airflow connection ID
    :return bool: If the connection exists, return True
    """
    BaseHook.get_connection(conn_id)
    return True



from __future__ import annotations

import random
import string
from typing import Any

from attr import define, field, fields_dict
from sqlalchemy import Column, MetaData

from astro.airflow.datasets import Dataset
from astro.databases import create_database
from astro.settings import OPENLINEAGE_EMIT_TEMP_TABLE_EVENT

MAX_TABLE_NAME_LENGTH = 62
TEMP_PREFIX = "_tmp"


@define
class Metadata:
    """
    Contains additional information to access a SQL Table, which is very likely optional and, in some cases, may
    be database-specific.

    :param schema: A schema name
    :param database: A database name
    """

    # This property is used by several databases, including: Postgres, Snowflake and BigQuery ("namespace")
    schema: str | None = None
    database: str | None = None

    def is_empty(self) -> bool:
        """Check if all the fields are None."""
        return all(getattr(self, field_name) is None for field_name in fields_dict(self.__class__))


@define(slots=False)
class BaseTable:
    """
    Base class that has information necessary to access a SQL Table. It is agnostic to the database type.
    If no name is given, it auto-generates a name for the Table and considers it temporary.

    Temporary tables are prefixed with the prefix TEMP_PREFIX.

    :param name: The name of the database table. If name not provided then it would create a temporary name
    :param conn_id: The Airflow connection id. This will be used to identify the right database type at the runtime
    :param metadata: A metadata object which will have database or schema name
    :param columns: columns which define the database table schema.
    :sphinx-autoapi-skip:
    """

    template_fields = ("name",)

    # TODO: discuss alternative names to this class, since it contains metadata as opposed to be the
    # SQL table itself
    # Some ideas: TableRef, TableMetadata, TableData, TableDataset
    name: str = field(default="")
    conn_id: str = field(default="")
    # Setting converter allows passing a dictionary to metadata arg
    metadata: Metadata = field(
        factory=Metadata,
        converter=lambda val: Metadata(**val) if isinstance(val, dict) else val,
    )
    columns: list[Column] = field(factory=list)
    temp: bool = field(default=False)

    def __attrs_post_init__(self) -> None:
        if not self.name:
            self.name = self._create_unique_table_name(TEMP_PREFIX + "_")
            self.temp = True
        if self.name.startswith(TEMP_PREFIX):
            self.temp = True

    # We need this method to pickle Table object, without this we cannot push/pull this object from xcom.
    def __getstate__(self):
        return self.__dict__

    def _create_unique_table_name(self, prefix: str = "") -> str:
        """
        If a table is instantiated without a name, create a unique table for it.
        This new name should be compatible with all supported databases.
        """
        schema_length = len((self.metadata and self.metadata.schema) or "") + 1
        prefix_length = len(prefix)

        unique_id = random.choice(string.ascii_lowercase) + "".join(
            random.choice(string.ascii_lowercase + string.digits)
            for _ in range(MAX_TABLE_NAME_LENGTH - schema_length - prefix_length)
        )
        if prefix:
            unique_id = f"{prefix}{unique_id}"

        return unique_id

    def create_similar_table(self) -> Table:
        """
        Create a new table with a unique name but with the same metadata.
        """
        return Table(  # type: ignore
            name=self._create_unique_table_name(),
            conn_id=self.conn_id,
            metadata=self.metadata,
        )

    @property
    def sqlalchemy_metadata(self) -> MetaData:
        """Return the Sqlalchemy metadata for the given table."""
        if self.metadata and self.metadata.schema:
            alchemy_metadata = MetaData(schema=self.metadata.schema)
        else:
            alchemy_metadata = MetaData()
        return alchemy_metadata

    @property
    def row_count(self) -> Any:
        """
        Return the row count of table.
        """
        db = create_database(self.conn_id)
        return db.row_count(self)

    @property
    def sql_type(self) -> Any:
        if self.conn_id:
            return create_database(self.conn_id).sql_type

    def to_json(self):
        return {
            "class": "Table",
            "name": self.name,
            "metadata": {
                "schema": self.metadata.schema,
                "database": self.metadata.database,
            },
            "temp": self.temp,
            "conn_id": self.conn_id,
        }

    @classmethod
    def from_json(cls, obj: dict):
        return Table(
            name=obj["name"],
            metadata=Metadata(**obj["metadata"]),
            temp=obj["temp"],
            conn_id=obj["conn_id"],
        )

    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        database = create_database(self.conn_id)
        return database.openlineage_dataset_name(table=self)

    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        database = create_database(self.conn_id)
        return database.openlineage_dataset_namespace()

    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        database = create_database(self.conn_id)
        return f"{database.openlineage_dataset_uri(table=self)}"

    def openlineage_emit_temp_table_event(self):
        """
        Based on airflow config ```OPENLINEAGE_EMIT_TEMP_TABLE_EVENT``` value and table type
        check whether to emit temp table event in openlineage or not
        """
        return (not isinstance(self, TempTable)) or (
            isinstance(self, TempTable) and OPENLINEAGE_EMIT_TEMP_TABLE_EVENT
        )


@define(slots=False)
class TempTable(BaseTable):
    """
    Internal class to represent a Temporary table

    :sphinx-autoapi-skip:
    """

    temp: bool = field(default=True)


@define(slots=False)
class Table(BaseTable, Dataset):
    """
    User-facing class that has information necessary to access a SQL Table. It is agnostic to the database type.
    If no name is given, it auto-generates a name for the Table and considers it temporary.

    Temporary tables are prefixed with the prefix TEMP_PREFIX.

    :param name: The name of the database table. If name not provided then it would create a temporary name
    :param conn_id: The Airflow connection id. This will be used to identify the right database type at the runtime
    :param metadata: A metadata object which will have database or schema name
    :param columns: columns which define the database table schema.
    """

    uri: str = field(init=False, eq=False)
    extra: dict | None = field(init=False, factory=dict)

    def __new__(cls, *args, **kwargs):
        name = kwargs.get("name") or args and args[0] or ""
        temp = kwargs.get("temp", False)
        if temp or (not name or name.startswith("_tmp")):
            return TempTable(*args, **kwargs)
        return super().__new__(cls)

    @uri.default
    def _path_to_dataset_uri(self) -> str:
        """Build a URI to be passed to Dataset obj introduced in Airflow 2.4"""
        from urllib.parse import urlencode, urlparse

        path = f"astro://{self.conn_id}@"
        db_extra = {"table": self.name}
        if self.metadata.schema:
            db_extra["schema"] = self.metadata.schema
        if self.metadata.database:
            db_extra["database"] = self.metadata.database
        parsed_url = urlparse(url=path)
        new_parsed_url = parsed_url._replace(query=urlencode(db_extra))
        return new_parsed_url.geturl()



from typing import Any, MutableMapping

# The ``airflow.utils.context.Context`` class was not available in Apache Airflow until 2.3.3. This class is
# typically used as the typing for the ``context`` arg in operators and sensors. However, using this class for
# typing outside of TYPE_CHECKING in modules sets an implicit, minimum requirement for Apache Airflow 2.2.3
# which is currently more recent than the current minimum requirement of Apache Airflow 2.2.0.
#
# TODO: Remove this once the repo has a minimum Apache Airflow requirement of 2.2.3+.
try:
    from airflow.utils.context import Context
except ModuleNotFoundError:  # pragma: no cover

    class Context(MutableMapping[str, Any]):  # type: ignore[no-redef] # skipcq PYL-W0223
        """Placeholder typing class for ``airflow.utils.context.Context``."""


__all__ = ["Context"]



from __future__ import annotations

from pathlib import Path
from typing import Any

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader
from jinja2.runtime import StrictUndefined


def render(template_file: Path, context: dict[str, Any], output_file: Path) -> None:
    """
    Render the template_file with context to a file.

    :param template_file: The path to the template file.
    :param context: The context to pass to the template file.
    :param output_file: The file to output the rendered version.
    """
    (
        Environment(
            loader=FileSystemLoader(Path(__file__).parent),
            undefined=StrictUndefined,
            autoescape=True,
            keep_trailing_newline=True,
        )
        .get_template(template_file.as_posix())
        .stream(**context)
        .dump(output_file.as_posix())
    )



from typing import TYPE_CHECKING

from astro.files.base import File, resolve_file_path_pattern  # noqa: F401 # skipcq: PY-W2000

if TYPE_CHECKING:
    from airflow.models.xcom_arg import XComArg


def get_file_list(path: str, conn_id: str, **kwargs) -> "XComArg":
    """
    List file path from a remote object store or the local filesystem based on the given path pattern.
    It is not advisable to fetch huge number of files since it would overload the XCOM and
    also, if you are using response of this method in dynamic task map ``expand`` method
    it can create huge number of parallel task

    Supported filesystem: :ref:`file_location`

    Supported file pattern: :ref:`file_pattern`

    :param conn_id: Airflow connection id.
        This will be used to identify the right Airflow hook at runtime to connect with storage services
    :param path: Path pattern to the file in the filesystem/Object stores
    """
    from astro.files.operators.files import ListFileOperator

    return ListFileOperator(
        path=path,
        conn_id=conn_id,
        **kwargs,
    ).output



from __future__ import annotations

import sys

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from cached_property import cached_property

if sys.version_info >= (3, 9):
    from functools import cache
else:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


__all__ = ["cache", "cached_property"]



from __future__ import annotations

import socket

from airflow.providers.sqlite.hooks.sqlite import SqliteHook
from sqlalchemy import MetaData as SqlaMetaData, create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.sql.schema import Table as SqlaTable

from astro.constants import MergeConflictStrategy
from astro.databases.base import BaseDatabase
from astro.options import LoadOptions
from astro.table import BaseTable, Metadata
from astro.utils.compat.functools import cached_property

DEFAULT_CONN_ID = SqliteHook.default_conn_name


class SqliteDatabase(BaseDatabase):
    """
    Handle interactions with Sqlite databases. If this class is successful, we should not have any Sqlite-specific
    logic in other parts of our code-base.
    """

    def __init__(
        self,
        conn_id: str = DEFAULT_CONN_ID,
        table: BaseTable | None = None,
        load_options: LoadOptions | None = None,
    ):
        super().__init__(conn_id)
        self.table = table
        self.load_options = load_options

    @property
    def sql_type(self) -> str:
        return "sqlite"

    @cached_property
    def hook(self) -> SqliteHook:
        """Retrieve Airflow hook to interface with the Sqlite database."""
        return SqliteHook(sqlite_conn_id=self.conn_id)

    @cached_property
    def sqlalchemy_engine(self) -> Engine:
        """Return SQAlchemy engine."""
        # Airflow uses sqlite3 library and not SqlAlchemy for SqliteHook
        # and it only uses the hostname directly.
        airflow_conn = self.hook.get_connection(self.conn_id)
        return create_engine(f"sqlite:///{airflow_conn.host}")

    @property
    def default_metadata(self) -> Metadata:
        """Since Sqlite does not use Metadata, we return an empty Metadata instances."""
        return Metadata()

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_table_qualified_name(table: BaseTable) -> str:
        """
        Return the table qualified name.

        :param table: The table we want to retrieve the qualified name for.
        """
        return str(table.name)

    def populate_table_metadata(self, table: BaseTable) -> BaseTable:
        """
        Since SQLite does not have a concept of databases or schemas, we just return the table as is,
        without any modifications.
        """
        table.conn_id = table.conn_id or self.conn_id
        return table

    def create_schema_if_needed(self, schema: str | None) -> None:
        """
        Since SQLite does not have schemas, we do not need to set a schema here.
        """

    def schema_exists(self, schema: str) -> bool:  # skipcq PYL-W0613,PYL-R0201
        """
        Check if a schema exists. We return false for sqlite since sqlite does not have schemas
        """
        return False

    @staticmethod
    def get_merge_initialization_query(parameters: tuple) -> str:
        """
        Handles database-specific logic to handle index for Sqlite.
        """
        return "CREATE UNIQUE INDEX merge_index ON {{table}}(%s)" % ",".join(parameters)  # skipcq PYL-C0209

    def merge_table(
        self,
        source_table: BaseTable,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        if_conflicts: MergeConflictStrategy = "exception",
    ) -> None:
        """
        Merge the source table rows into a destination table.
        The argument `if_conflicts` allows the user to define how to handle conflicts.

        :param source_table: Contains the rows to be merged to the target_table
        :param target_table: Contains the destination table in which the rows will be merged
        :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
        :param target_conflict_columns: List of cols where we expect to have a conflict while combining
        :param if_conflicts: The strategy to be applied if there are conflicts.
        """
        statement = "INSERT INTO {main_table} ({target_columns}) SELECT {append_columns} FROM {source_table} Where true"
        if if_conflicts == "ignore":
            statement += " ON CONFLICT ({merge_keys}) DO NOTHING"
        elif if_conflicts == "update":
            statement += " ON CONFLICT ({merge_keys}) DO UPDATE SET {update_statements}"

        append_column_names = list(source_to_target_columns_map.keys())
        target_column_names = list(source_to_target_columns_map.values())
        update_statements = [f"{col_name}=EXCLUDED.{col_name}" for col_name in target_column_names]

        query = statement.format(
            target_columns=",".join(target_column_names),
            main_table=target_table.name,
            append_columns=",".join(append_column_names),
            source_table=source_table.name,
            update_statements=",".join(update_statements),
            merge_keys=",".join(list(target_conflict_columns)),
        )

        self.run_sql(sql=query)

    def get_sqla_table(self, table: BaseTable) -> SqlaTable:
        """
        Return SQLAlchemy table instance

        :param table: Astro Table to be converted to SQLAlchemy table instance
        """
        return SqlaTable(table.name, SqlaMetaData(), autoload_with=self.sqlalchemy_engine)

    def openlineage_dataset_name(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: /tmp/local.db.table_name
        """
        conn = self.hook.get_connection(self.conn_id)
        return f"{conn.host}.{table.name}"

    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: file://127.0.0.1:22
        """
        conn = self.hook.get_connection(self.conn_id)
        port = conn.port or 22
        return f"file://{socket.gethostbyname(socket.gethostname())}:{port}"

    def openlineage_dataset_uri(self, table: BaseTable) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return f"{self.openlineage_dataset_namespace()}{self.openlineage_dataset_name(table=table)}"



from __future__ import annotations

import json
import os
from urllib.parse import urlparse, urlunparse

import smart_open
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from google.api_core.exceptions import NotFound

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation


def _pull_credentials_from_json_dict(creds_json: dict) -> dict:
    expected_values = {"client_email", "project_id", "private_key", "private_key_id"}
    missing_values = expected_values - set(creds_json.keys())
    if missing_values:
        raise ValueError(f"Error retrieving GCP credentials. missing key(s): {','.join(missing_values)}")
    return {
        "spark.hadoop.google.cloud.auth.service.account.enable": "true",
        "spark.hadoop.fs.gs.auth.service.account.email": creds_json["client_email"],
        "spark.hadoop.fs.gs.project.id": creds_json["project_id"],
        "spark.hadoop.fs.gs.auth.service.account.private.key": creds_json["private_key"],
        "spark.hadoop.fs.gs.auth.service.account.private.key.id": creds_json["private_key_id"],
    }


def _pull_credentials_from_keypath(creds_dict: dict) -> dict:
    with open(creds_dict["key_path"]) as f:
        return _pull_credentials_from_json_dict(json.loads(f.read()))


class GCSLocation(BaseFileLocation):
    """Handler GS object store operations"""

    location_type = FileLocation.GS
    # TODO: Restrict the supported conn_type to only GCSHook.conn_type
    supported_conn_type = {GCSHook.conn_type, "gcpbigquery", "bigquery"}

    @property
    def hook(self) -> GCSHook:
        return GCSHook(gcp_conn_id=self.conn_id) if self.conn_id else GCSHook()

    @property
    def transport_params(self) -> dict:
        """get GCS credentials for storage"""
        client = self.hook.get_conn()
        return {"client": client}

    @property
    def paths(self) -> list[str]:
        """Resolve GS file paths with prefix"""
        url = urlparse(self.path)
        bucket_name = url.netloc
        prefix = url.path[1:]
        prefixes = self.hook.list(bucket_name=bucket_name, prefix=prefix)
        paths = [urlunparse((url.scheme, url.netloc, keys, "", "", "")) for keys in prefixes]
        return paths

    @property
    def size(self) -> int:
        """Return file size for GCS location"""
        url = urlparse(self.path)
        bucket_name = url.netloc
        object_name = url.path
        if object_name.startswith("/"):
            object_name = object_name[1:]
        return int(self.hook.get_size(bucket_name=bucket_name, object_name=object_name))

    def databricks_auth_settings(self) -> dict:
        """
        Required settings to upload this file into databricks. Only needed for cloud storage systems
        like S3
        :return: A dictionary of settings keys to settings values
        """
        creds_dict = self.hook.get_connection(self.conn_id).extra_dejson
        if creds_dict.get("key_path"):
            return _pull_credentials_from_keypath(creds_dict=creds_dict)
        elif creds_dict.get("keyfile_dict"):
            return _pull_credentials_from_json_dict(creds_dict.get("keyfile_dict"))
        elif os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
            return _pull_credentials_from_keypath(
                {"key_path": os.getenv("GOOGLE_APPLICATION_CREDENTIALS", "")}
            )
        else:
            raise ValueError(
                "Error: to pull credentials from GCP We either need a keyfile or a keyfile_dict "
                "to retrieve credentials"
            )

    def exists(self) -> bool:
        """Check if the file exists or not"""
        try:
            with smart_open.open(self.smartopen_uri, mode="r", transport_params=self.transport_params):
                return True
        except (OSError, NotFound):
            return False

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        parsed_url = urlparse(self.path)
        return f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return urlparse(self.path).path



from __future__ import annotations

import attr

from astro.constants import ExportExistsStrategy, FileType, MergeConflictStrategy
from astro.lineage import BaseFacet
from astro.table import Column, Metadata


@attr.define
class InputFileFacet(BaseFacet):
    """
    Facet that represents input file for load file

    :param filepath: path of each file.
    :param file_size: size of the file.
    :param file_type: type of the file.
    """

    filepath: str
    file_size: int | None
    file_type: FileType
    description: str | None = None


@attr.define
class ExportFileFacet(BaseFacet):
    """
    Facet that represents output file for export file

    :param filepath: path of each file.
    :param file_size: size of the file.
    :param file_type: type of the file.
    :param if_exists: Overwrite file if exists.
    """

    filepath: str
    file_size: int | None
    file_type: FileType
    if_exists: ExportExistsStrategy


@attr.define
class InputFileDatasetFacet(BaseFacet):
    """
    Facet that represents input file dataset Facets for load file

    :param file_size: size of the file.
    :param number_of_files: number of files to be loaded.
    :param file_type: type of the file.
    :param description: description of the dataset.
    :param is_pattern: True when file path is a pattern(eg. ``s3://bucket/folder`` or ``/folder/sample_*`` etc).
    :param files: list of filepaths to be loaded from dataset.
    """

    number_of_files: int | None
    description: str | None = None
    is_pattern: bool = False
    files: list[InputFileFacet] = []


@attr.define
class OutputDatabaseDatasetFacet(BaseFacet):
    """
    Facet that represents output database dataset Facets for load file

    :param metadata: metadata of the table.
    :param columns: columns defined in table
    :param schema: schema used.
    :param used_native_path: use native support for data transfer if available on the destination.
    :param enabled_native_fallback: use enable_native_fallback=True to fall back to default transfer.
    :param native_support_arguments: args to be used by method involved in native support flow
    :param description: description of the dataset.
    """

    metadata: Metadata
    columns: list[Column]
    schema: str | None
    used_native_path: bool
    enabled_native_fallback: bool | None = False
    native_support_arguments: dict | None = None
    description: str | None = None

    _do_not_redact = ["metadata", "columns"]


@attr.define
class TableDatasetFacet(BaseFacet):
    """
    Facets that represent may be available for Table object

    :param table_name: Name of the table
    :param columns: columns defined in table
    :param source_table_rows: Total rows in source table
    :param metadata: metadata of the table.
    """

    table_name: str
    columns: list[str] | tuple[str] | dict[str, str] | None = None
    source_table_rows: int = 0
    metadata: Metadata | None = None

    _do_not_redact = ["metadata", "columns"]


@attr.define
class SourceTableMergeDatasetFacet(BaseFacet):
    """
    Facets that represent may be available for Source Table object in MergeOperator

    :param table_name: Name of the table
    :param if_conflicts: The strategy to be applied if there are conflicts.
    :param columns: columns defined in table
    :param source_table_rows: Total rows in source table
    :param metadata: metadata of the table.
    """

    table_name: str
    if_conflicts: MergeConflictStrategy
    columns: list[str] | tuple[str] | dict[str, str] | None = None
    source_table_rows: int = 0
    metadata: Metadata | None = None

    _do_not_redact = ["metadata", "columns"]


@attr.define
class TargetTableMergeDatasetFacet(BaseFacet):
    """
    Facets that represent may be available for Target Table object in Merge Operator

    :param table_name: Name of the table
    :param target_conflict_columns: List of cols where we expect to have a conflict while combining
    :param columns: columns defined in table
    :param metadata: metadata of the table.
    """

    table_name: str
    target_conflict_columns: list[str]
    columns: list[str] | tuple[str] | dict[str, str] | None = None
    metadata: Metadata | None = None

    _do_not_redact = ["metadata", "columns"]



import fileinput
import logging
import pathlib
import time
from pathlib import Path
from typing import Dict

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from databricks_cli.dbfs.api import DbfsApi, DbfsPath
from databricks_cli.runs.api import RunsApi
from databricks_cli.sdk.api_client import ApiClient
from databricks_cli.secrets.api import SecretApi
from requests.exceptions import HTTPError

from astro.databases.databricks.load_file.load_file_python_code_generator import render
from astro.databases.databricks.load_options import DeltaLoadOptions

cwd = pathlib.Path(__file__).parent
log = logging.getLogger(__name__)


def delete_secret_scope(scope_name: str, api_client: ApiClient) -> None:
    """
    Delete the scope we created to prevent littering the databricks secret store with one-time scopes.

    :param scope_name: name of scope to delete
    :param api_client: populated databricks client
    """
    secrets = SecretApi(api_client)
    try:
        secrets.delete_scope(scope_name)
    except HTTPError as http_error:
        # We don't care if the resource doesn't exist since we're trying to delete it
        if not http_error.response.json().get("error_code", "") == "RESOURCE_DOES_NOT_EXIST":
            raise http_error


def create_secrets(scope_name: str, filesystem_secrets: Dict[str, str], api_client: ApiClient) -> None:
    """
    Uploads secrets to a scope
    Before we can transfer data from external file sources (s3, GCS, etc.) we first need to upload the relevant
    secrets to databricks, so we can use them in the autoloader config. This allows us to perform ad-hoc queries
    that are not dependent on existing settings.

    :param scope_name: the name of the secret scope. placing all secrets in a single scope makes them easy to
        delete later
    :param filesystem_secrets: a dictionary of k,v secrets where the key is the secret name and the value is
        the secret value
    :param api_client: The databricks API client that has all necessary credentials
    """
    secrets = SecretApi(api_client)
    secrets.create_scope(
        scope=scope_name,
        initial_manage_principal=None,
        backend_azure_keyvault=None,
        scope_backend_type=None,
    )

    for k, v in filesystem_secrets.items():
        # Add "astro_sdk_" to the front, so we know which secrets are hadoop settings
        secret_key = "astro_sdk_" + k
        secrets.put_secret(scope=scope_name, key=secret_key, string_value=v, bytes_value=None)


def generate_file(
    data_source_path: str,
    table_name: str,
    source_type: str,
    output_file_path: Path,
    load_options: DeltaLoadOptions,
    file_type: str = "",
):
    """
    In order to run autoloader jobs in databricks, we need to generate a python file that creates a pyspark job.
    This function uses jinja templating to generate a file with all necessary user inputs

    :param data_source_path: The path where the data we're injesting lives (e.g. the s3 bucket URL)
    :param table_name: The delta table we would like to push the data into
    :param source_type: the source location (e.g. s3)
    :param load_options: any additional load options the user would like to inject into their job
    :param output_file_path: where the generated file should be placed.
    :param file_type: when using the COPY INTO command, Spark requires explicitly stating the data type you are loading.
        For individual files we can infer the file type, but if you are loading a directory please explicitly state
        the file type in the File object.
    :return: output file path
    """
    render(
        Path("jinja_templates/load_file_to_delta.py.jinja2"),
        {
            "source_type": source_type,
            "table_name": table_name,
            "load_options": load_options,
            "input_path": data_source_path,
            "file_type": file_type.upper(),
            "if_exists": load_options.if_exists,
        },
        output_file_path,
    )

    # Since we had to use jinja2 safe strings, we want to inline replace \' with '
    # This is not a security concern now that jinja2 is no longer running.
    for line in fileinput.input(output_file_path, inplace=True):
        print(line.replace("\\'", "'"), end="")
    log.debug("Created file %s", output_file_path)
    return output_file_path


def load_file_to_dbfs(local_file_path: Path, file_name: str, api_client: ApiClient) -> Path:
    # TODO we should allow arbitrary dbfs paths set by env vars/users
    """Load a file into DBFS. Used to move a python file into DBFS, so we can run the jobs as pyspark jobs.
    Currently saves the file to dbfs:/mnt/pyscripts/, but will eventually support more locations.

    :param local_file_path: path of the file to upload
    :param api_client: Databricks API client
    :return: path to the file in DBFS

    """
    log.debug("loading file %s", str(local_file_path))
    dbfs = DbfsApi(api_client=api_client)
    file_path = DbfsPath("dbfs:/mnt/pyscripts/")
    dbfs.mkdirs(file_path)
    dbfs.delete(file_path.join(file_name), recursive=False)
    dbfs.put_file(src_path=str(local_file_path), dbfs_path=file_path.join(file_name), overwrite=True)
    return Path("dbfs:/mnt/pyscripts") / file_name


def create_and_run_job(
    api_client,
    file_to_run: str,
    databricks_job_name: str,
    existing_cluster_id=None,
    new_cluster_specs=None,
) -> None:
    """Creates a databricks job and runs it to completion.

    :param databricks_job_name: Name of the job to submit
    :param file_to_run: path to file we will run with the job
    :param api_client: databricks API client
    :param existing_cluster_id: If you want to run this job on an existing cluster, you can give the cluster ID
    :param new_cluster_specs: If you want to run this job on a new cluster, you can give the cluster specs
        as a python dictionary.

    """
    runs_api = RunsApi(api_client=api_client)
    json_info = {
        "name": databricks_job_name,
        "spark_python_task": {"python_file": file_to_run},
    }

    if new_cluster_specs:
        log.debug("creating new spark cluster with spec %s", new_cluster_specs)
        json_info["new_cluster"] = new_cluster_specs
    elif existing_cluster_id:
        log.debug("Using existing cluster %s", existing_cluster_id)
        json_info["existing_cluster_id"] = existing_cluster_id
    else:
        raise AirflowException("Error, must have a new cluster spec or an existing cluster ID for spark job")

    run_id = runs_api.submit_run(json=json_info)["run_id"]

    ping_interval = int(conf.get("astro_sdk", "databricks_ping_interval", fallback=5))
    while runs_api.get_run(run_id)["state"]["life_cycle_state"] == "PENDING":
        log.info("job pending")
        time.sleep(ping_interval)

    while runs_api.get_run(run_id)["state"]["life_cycle_state"] == "RUNNING":
        log.info("job running")
        time.sleep(ping_interval)
    final_job_state = runs_api.get_run(run_id)
    final_state = final_job_state["state"]["result_state"]
    log.info("final state: %s", {final_state})
    if final_state == "FAILED":
        raise AirflowException(f"Databricks job failed. Job info {final_job_state}")



from __future__ import annotations

from airflow.configuration import conf
from attr import define, field

from astro.constants import DatabricksLoadMode, LoadExistStrategy
from astro.options import LoadOptions


@define
class DeltaLoadOptions(LoadOptions):
    """
    Helper class for rendering options into COPY_INTO Databricks SQL statements

    :param copy_into_format_options: File format options for COPY INTO. Read more at:
        https://docs.databricks.com/sql/language-manual/delta-copy-into.html#format-options
    :param copy_into_copy_options: Options to control the operation of the COPY INTO command. Read more at:
        https://docs.databricks.com/sql/language-manual/delta-copy-into.html
    :param existing_cluster_id: Existing cluster ID on databricks.
    :param new_cluster_spec: New cluster specifications.
    :param if_exists: default override an existing Table. Options: fail, replace, append
    :param secret_scope: Secrets scope
    :param load_secrets: Defaults to False.
    :param load_mode: defaults to autoloader. Options: autoloader, COPY INTO
    :param autoloader_load_options: Options for load using autoloader. Read more at:
        https://docs.databricks.com/ingestion/auto-loader/options.html
    """

    copy_into_format_options: dict = field(init=True, factory=dict)
    copy_into_copy_options: dict = field(init=True, factory=dict)
    existing_cluster_id: str = field()
    new_cluster_spec: dict = field(init=True, factory=dict)
    if_exists: LoadExistStrategy = "replace"
    secret_scope: str = "astro-sdk-secrets"
    load_secrets: bool = False
    load_mode: DatabricksLoadMode = DatabricksLoadMode.AUTOLOADER
    autoloader_load_options: dict = field(factory=dict)

    @existing_cluster_id.default
    def get_existing_cluster_id(self):
        return conf.get("astro_sdk", "databricks_cluster_id")

    def empty(self):
        return not (
            self.copy_into_copy_options
            and self.copy_into_format_options
            and self.new_cluster_spec
            and self.existing_cluster_id
        )

    def convert_format_options_to_string(self):
        return ",".join([f"'{k}' = '{v}'" for k, v in self.copy_into_format_options.items()])

    def convert_copy_options_to_string(self):
        return ",".join([f"'{k}' = '{v}'" for k, v in self.copy_into_copy_options.items()])

    @classmethod
    def get_default_delta_options(cls):
        return cls(
            copy_into_format_options={"header": "true", "inferSchema": "true"},
            copy_into_copy_options={"mergeSchema": "true"},
            load_secrets=conf.get("astro_sdk", "load_storage_configs_to_databricks", fallback=True),
        )



from typing import Any, Dict, Optional, Union

import pandas
from airflow import AirflowException
from airflow.decorators.base import get_unique_task_id
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.providers.common.sql.operators.sql import SQLColumnCheckOperator

from astro.databases import create_database
from astro.table import BaseTable
from astro.utils.compat.typing import Context


class ColumnCheckOperator(SQLColumnCheckOperator):
    """
    Performs one or more of the templated checks in the column_checks dictionary.
    Checks are performed on a per-column basis specified by the column_mapping.
    Each check can take one or more of the following options:
    - equal_to: an exact value to equal, cannot be used with other comparison options
    - greater_than: value that result should be strictly greater than
    - less_than: value that results should be strictly less than
    - geq_to: value that results should be greater than or equal to
    - leq_to: value that results should be less than or equal to
    - tolerance: the percentage that the result may be off from the expected value

    :param dataset: the table or dataframe to run checks on
    :param column_mapping: the dictionary of columns and their associated checks, e.g.

    .. code-block:: python

        {
            "col_name": {
                "null_check": {
                    "equal_to": 0,
                },
                "min": {
                    "greater_than": 5,
                    "leq_to": 10,
                    "tolerance": 0.2,
                },
                "max": {"less_than": 1000, "geq_to": 10, "tolerance": 0.01},
            }
        }
    """

    def __init__(
        self,
        dataset: Union[BaseTable, pandas.DataFrame],
        column_mapping: Dict[str, Dict[str, Any]],
        partition_clause: Optional[str] = None,
        task_id: Optional[str] = None,
        **kwargs,
    ):
        for checks in column_mapping.values():
            for check, check_values in checks.items():
                self._column_mapping_validation(check, check_values)

        self.dataset = dataset
        self.column_mapping = column_mapping
        self.partition_clause = partition_clause
        self.kwargs = kwargs
        self.df = None

        dataset_qualified_name = ""
        dataset_conn_id = ""

        if isinstance(dataset, BaseTable):
            db = create_database(conn_id=self.dataset.conn_id)  # type: ignore
            self.conn_id = self.dataset.conn_id
            dataset_qualified_name = db.get_table_qualified_name(table=self.dataset)
            dataset_conn_id = dataset.conn_id

        super().__init__(
            table=dataset_qualified_name,
            column_mapping=self.column_mapping,
            partition_clause=self.partition_clause,
            conn_id=dataset_conn_id,
            task_id=task_id if task_id is not None else get_unique_task_id("check_column"),
        )

    def get_db_hook(self) -> DbApiHook:
        """
        Get the database hook for the connection.

        :return: the database hook object.
        """
        db = create_database(conn_id=self.conn_id)
        if db.sql_type == "bigquery":
            return db.hook
        return super().get_db_hook()

    def execute(self, context: "Context"):
        if isinstance(self.dataset, BaseTable):
            return super().execute(context=context)
        elif isinstance(self.dataset, pandas.DataFrame):
            self.df = self.dataset
        else:
            raise ValueError("dataset can only be of type pandas.dataframe | Table object")

        self.process_checks()

    def get_check_result(self, check_name: str, column_name: str):
        """
        Get the check method results post validating the dataframe
        """
        if self.df is not None and column_name in self.df.columns:
            column_checks = {
                "null_check": lambda column: column.isna().sum(),
                "distinct_check": lambda column: len(column.unique()),
                "unique_check": lambda column: len(column) - len(column.unique()),
                "min": lambda column: column.min(),
                "max": lambda column: column.max(),
            }
            return column_checks[check_name](column=self.df[column_name])
        if self.df is None:
            raise ValueError("Dataframe is None")
        if column_name not in self.df.columns:
            raise ValueError(f"Dataframe doesn't have column {column_name}")

    def process_checks(self):
        """
        Process all the checks and print the result or raise an exception in the event of failed checks
        """
        failed_tests = []
        passed_tests = []

        # Iterating over columns
        for column in self.column_mapping:
            checks = self.column_mapping[column]

            # Iterating over checks
            for check_key, check_val in checks.items():
                tolerance = check_val.get("tolerance")
                result = self.get_check_result(check_key, column_name=column)
                check_val["result"] = result
                check_val["success"] = self._get_match(check_val, result, tolerance)
                failed_tests.extend(_get_failed_checks(checks, column))
                passed_tests.extend(_get_success_checks(checks, column))

        if len(failed_tests) > 0:
            raise AirflowException(f"The following tests have failed: \n{''.join(failed_tests)}")
        if len(passed_tests) > 0:
            print(f"The following tests have passed: \n{''.join(passed_tests)}")


def _get_failed_checks(checks, col=None):
    return [
        f"{get_checks_string(checks, col)} {check_values}\n"
        for check, check_values in checks.items()
        if not check_values["success"]
    ]


def _get_success_checks(checks, col=None):
    return [
        f"{get_checks_string(checks, col)} {check_values}\n"
        for check, check_values in checks.items()
        if check_values["success"]
    ]


def get_checks_string(check, col):
    if col:
        return f"Column: {col}\nCheck: {check},\nCheck Values:"
    return f"\tCheck: {check},\n\tCheck Values:"


def check_column(
    dataset: Union[BaseTable, pandas.DataFrame],
    column_mapping: Dict[str, Dict[str, Any]],
    partition_clause: Optional[str] = None,
    task_id: Optional[str] = None,
    **kwargs,
) -> ColumnCheckOperator:
    """
    Performs one or more of the templated checks in the column_checks dictionary.
    Checks are performed on a per-column basis specified by the column_mapping.
    Each check can take one or more of the following options:
    - equal_to: an exact value to equal, cannot be used with other comparison options
    - greater_than: value that result should be strictly greater than
    - less_than: value that results should be strictly less than
    - geq_to: value that results should be greater than or equal to
    - leq_to: value that results should be less than or equal to
    - tolerance: the percentage that the result may be off from the expected value

    :param dataset: dataframe or BaseTable that has to be validated
    :param column_mapping: the dictionary of columns and their associated checks, e.g.

    .. code-block:: python

        {
            "col_name": {
                "null_check": {
                    "equal_to": 0,
                },
                "min": {
                    "greater_than": 5,
                    "leq_to": 10,
                    "tolerance": 0.2,
                },
                "max": {"less_than": 1000, "geq_to": 10, "tolerance": 0.01},
            }
        }
    """
    return ColumnCheckOperator(
        dataset=dataset,
        column_mapping=column_mapping,
        partition_clause=partition_clause,
        kwargs=kwargs,
        task_id=task_id,
    )



import importlib
from pathlib import Path
from typing import Optional

from astro.constants import FileLocation
from astro.files.locations.base import BaseFileLocation
from astro.options import LoadOptionsList
from astro.utils.path import get_class_name, get_dict_with_module_names_to_dot_notations

DEFAULT_FILE_SCHEME_TO_MODULE_PATH = get_dict_with_module_names_to_dot_notations(Path(__file__))
DEFAULT_FILE_SCHEME_TO_MODULE_PATH["https"] = DEFAULT_FILE_SCHEME_TO_MODULE_PATH["http"]
DEFAULT_FILE_SCHEME_TO_MODULE_PATH["gs"] = DEFAULT_FILE_SCHEME_TO_MODULE_PATH["gcs"]
DEFAULT_FILE_SCHEME_TO_MODULE_PATH["wasbs"] = DEFAULT_FILE_SCHEME_TO_MODULE_PATH["wasb"]


def create_file_location(
    path: str, conn_id: Optional[str] = None, load_options_list: Optional[LoadOptionsList] = None
) -> BaseFileLocation:
    """
    Location factory method to generate location class

    :param path: Path to a file in the filesystem/Object stores
    :param conn_id: Airflow connection ID
    """
    filetype: FileLocation = BaseFileLocation.get_location_type(path)
    module_path = DEFAULT_FILE_SCHEME_TO_MODULE_PATH[filetype.value]
    module_ref = importlib.import_module(module_path)
    class_name = get_class_name(module_ref)
    location: BaseFileLocation = getattr(module_ref, class_name)(
        path,
        conn_id,
        load_options=load_options_list and load_options_list.get(getattr(module_ref, class_name)),
    )
    return location
